
(* for building this file:
   make version.sml
   run   poly < mk_observers.sml > log
   get the description of observers in log and pack each into the relevant newXXX function below
   make sure to replace dummy constant bounds in log by d1 and d2, fill probes, update roots, rename names ...
*)


(* ugly but seems to work: *)
(* bound difference must be exact, even though bound are represented by SML reals
    e.g. we MUST have 0.44 - 0.1 = 0.34 rather than 0.339999.... obtained with Real.- ! *)
fun bdiff d2 d1 = Option.valOf (Real.fromString (Real.toString (d2-d1)));


structure Observers = struct

open Fiacre;

(* returns observed mains from an observable *)
fun roots p =
    let fun recroots pb rl =
	    case pb of
		StateProbe ((root,_),s) => makeset (root :: rl)
	      | ValueProbe ((root,_),p) => makeset (root :: rl)
	      | EventProbe ((root,_),p) => makeset (root :: rl)
	      | TagProbe ((root,_),p)   => makeset (root :: rl)
	      | EnterProbe pb      => recroots pb rl
	      | LeaveProbe pb      => recroots pb rl
	      | AndProbe (pb1,pb2) => recroots pb2 (recroots pb1 rl)
	      | OrProbe (pb1,pb2)  => recroots pb2 (recroots pb1 rl)
	      | ImpProbe (pb1,pb2) => recroots pb2 (recroots pb1 rl)
	      | NotProbe pb        => recroots pb rl
	      | ParenProbe pb      => recroots pb rl
	      (* normally removed by typer: *)
	      | JuxtProbe _        => failwith "unexpected JuxtProbe in convProbe"
	      | ThatProbe _        => failwith "unexpected ThatProbe in convProbe"
    in recroots p []
    end;


(* counter for generating unique observers *)

val nextobs = let val X = ref 0 in fn () => (X := 1 + !X; !X) end;

(* -- Observer generators -- *)
(* each mk_XXX returns a list of declarations and a property *)

(*
process presentAfter [E1, E2 : sync] is
   states idle, start, watch, error, stop
   from idle
      E2 ;
      to start
   from start
      wait [d1,d1];
      to watch
   from watch
      select
	 E1 ; to stop
      unless
	 wait ]d2−d1,...[;
	 to error
      end
component PresentAfter is
   port E1 : sync in [0, 0] is ...,
        E2 : sync in [0, 0] is ...
   par presentAfter [E1, E2] end
PresentAfter > observed...
property present_after is ltl [] - (PresentAfter/1/state error)
*)
     
fun newPresentAfter (e1,e2,SOME (Closed d1,Closed d2)) =
    let val no = Int.toString (nextobs ())
	fun ` s = "_" ^ s ^ "_" ^ no
    in ([ProcessDecl
	     ((3, 0), ((3, 8), `"presentAfter"), [([((3, 22), "E1"), ((3, 26), "E2")], ([IN, OUT], []))], [],
	      [((4, 10), "idle"), ((4, 16), "start"), ((4, 23), "watch"), ((4, 30), "error"), ((4, 37), "stop")], [], NONE,
	      [((5, 3), ((5, 8), "idle"),
		StSequence ((6, 8), StSignal (((6, 6), "E2"), ref Local, ref ([IN, OUT], [])), StTo ((7, 6), "start"))),
	       ((8, 3), ((8, 8), "start"),
		StSequence ((9, 16), StWait ((9, 6), Closed d1, Closed d1), StTo ((10, 6), "watch"))),
	       ((11, 3), ((11, 8), "watch"),
		StSelect ([[((14, 6), StSequence ((13, 5), StSignal (((13, 2), "E1"), ref Local, ref ([IN, OUT], [])), StTo ((13, 7), "stop")))],
			   [((12, 6), StSequence ((15, 14), StWait ((15, 2), Open (bdiff d2 d1), Infinite), StTo ((16, 2), "error")))]],
			  (17, 6)))]),
	 ComponentDecl
	     ((19, 0), ((19, 10), `"PresentAfter"), [], [], [],
	      [(([((20, 8), "E1")], ([IN, OUT], [])), SOME e1, (Closed 0.0, Closed 0.0)),
	       (([((21, 8), "E2")], ([IN, OUT], [])), SOME e2, (Closed 0.0, Closed 0.0))],
	      [], NONE,
	      ParComp ((22, 3), SomePorts [], [(SomePorts [], InstanceComp (((22, 7), `"presentAfter"), ["E1", "E2"], []))])),
	 Main
	     (((24, 0), `"PresentAfter"), roots (OrProbe(e1,e2))),
	 PropertyDecl
	     ((26, 8), ((26, 17), `"present_after"),
	      LtlProp (BoxProp  (NotProp  (ProbeProp (StateProbe ((`"PresentAfter", [1]), "error"))))))],
	VarProp (L,`"present_after"))
    end

  | newPresentAfter _ =
    comperror "unsupported time interval in present .. after ... within ..., sorry";



(*
process presentBefore1 [E1 : sync] (&flag : bool, &foundE2 : bool) is
   states idle, start, watch, error
   from idle
      E1;
      to start
   from start
      wait [d1, d1]; flag := true;
      to watch
   from watch
      wait ]d2−d1,...[;
      flag := false;
      on not foundE2;
      to error
process presentBefore2 [E2 : sync] (&foundE2 : bool, &flag : bool) is
   states idle, stop
   from idle
      E2 ;
      foundE2 := true;
      on flag;
      to stop
component PresentBefore is
   var flag : bool := false,
       foundE2 : bool := false
   port E1 : sync in [0,0] is ...,
        E2 : sync in [0,0] is ...
   par
      presentBefore1 [E1] (&flag, &foundE2)
   || presentBefore2 [E2] (&foundE2, &flag)
   end
PresentBefore > ....
property present_before is ltl <> (PresentBefore/1/value foundE2) => [] - (PresentBefore/1/state error)
 *)

fun newPresentBefore (e1,e2,SOME (Closed d1,Closed d2)) =
    let val no = Int.toString (nextobs ())
	fun ` s = "_" ^ s ^ "_" ^ no
    in ([ProcessDecl
	     ((32, 0), ((32, 8), `"presentBefore1"), [([((32, 24), "E1")], ([IN, OUT], []))],
	      [([RefArg ((32, 37), "flag")], [], BoolType), ([RefArg ((32, 51), "foundE2")], [], BoolType)],
	      [((33, 10), "idle"), ((33, 16), "start"), ((33, 23), "watch"), ((33, 30), "error")], [], NONE,
	      [((34, 3), ((34, 8), "idle"),
		StSequence ((35, 8), StSignal (((35, 6), "E1"), ref Local, ref ([IN, OUT], [])), StTo ((36, 6), "start"))),
	       ((37, 3), ((37, 8), "start"),
		StSequence ((38, 16), StWait ((38, 6), Closed d1, Closed d1),
			    StSequence ((38, 30), StAssign ((38, 23), [IdentExp ("flag", ref (VarKind, BoolType))], [BoolExp true], ref true),
					StTo ((39, 6), "watch")))),
	       ((40, 3), ((40, 8), "watch"),
		StSequence ((41, 18), StWait ((41, 6), Open (bdiff d2 d1), Infinite),
			    StSequence ((42, 19), StAssign ((42, 11), [IdentExp ("flag", ref (VarKind, BoolType))], [BoolExp false], ref true),
					StSequence  ((43, 20), StOn ((43, 6), UnopExp (NOT, ref BoolType, IdentExp ("foundE2", ref (VarKind, BoolType)))),
						     StTo ((44, 6), "error")))))]),
	 ProcessDecl
	     ((46, 0), ((46, 8), `"presentBefore2"), [([((46, 24), "E2")], ([IN, OUT], []))],
	      [([RefArg ((46, 37), "foundE2")], [], BoolType), ([RefArg ((46, 54), "flag")], [], BoolType)],
	      [((47, 10), "idle"), ((47, 16), "stop")], [], NONE,
	      [((48, 3), ((48, 8), "idle"),
		StSequence ((49, 9), StSignal (((49, 6), "E2"), ref Local, ref ([IN, OUT], [])),
			    StSequence ((50, 21), StAssign ((50, 14), [IdentExp ("foundE2", ref (VarKind, BoolType))], [BoolExp true], ref true),
					StSequence ((51, 13), StOn ((51, 6), IdentExp ("flag", ref (VarKind, BoolType))),
						    StTo ((52, 6), "stop")))))]),
	 ComponentDecl
	     ((54, 0), ((54, 10), `"PresentBefore"), [], [],
	      [([((55, 7), "flag")], BoolType, SOME (BoolExp false)), ([((56, 7), "foundE2")], BoolType, SOME (BoolExp false))],
	      [(([((57, 8), "E1")], ([IN, OUT], [])), SOME e1, (Closed 0.0, Closed 0.0)),
	       (([((58, 8), "E2")], ([IN, OUT], [])), SOME e2, (Closed 0.0, Closed 0.0))],
	      [], NONE,
	      ParComp
		  ((59, 3), SomePorts [],
		   [(SomePorts [], InstanceComp (((60, 6), `"presentBefore1"), ["E1"],
						 [RefExp ("flag", ref BoolType), RefExp ("foundE2", ref BoolType)])),
		    (SomePorts [], InstanceComp (((61, 6), `"presentBefore2"), ["E2"],
						 [RefExp ("foundE2", ref BoolType), RefExp ("flag", ref BoolType)]))])),
	 Main
	     (((64, 0), `"PresentBefore"), roots (OrProbe(e1,e2))),
	 PropertyDecl
	     ((66, 8), ((66, 17), `"present_before"),
	      LtlProp (ImpProp (DiaProp (ProbeProp (ValueProbe ((`"PresentBefore", [1]), IdentExp ("foundE2", ref (VarKind, BoolType))))),
				BoxProp (NotProp (ProbeProp (StateProbe ((`"PresentBefore", [1]), "error")))))))],
	VarProp (L,`"present_before"))
    end

  | newPresentBefore (e1,e2,_) =
    comperror "unsupported time interval in present .. before ... within ..., sorry";


(*
process presentWithin [E1 : sync] is
   states idle, watch, error, stop
   from idle
      wait [d1,d1];
      to watch
   from watch
      select
	 E1; to stop
      unless
	 wait ]d2−d1,...[;
	 to error
      end
component PresentWithin is
   port E1 : sync in [0,0] is ...
   par presentWithin [E1] end
PresentWithin > ...
property present_within is ltl [] - (PresentWithin/1/state error)
*)

fun newPresentWithin (e,SOME (Closed d1,Closed d2)) =
    let val no = Int.toString (nextobs ())
	fun ` s = "_" ^ s ^ "_" ^ no
    in ([ProcessDecl
	     ((71, 0), ((71, 8), `"presentWithin"), [([((71, 23), "E1")], ([IN, OUT], []))], [],
	      [((72, 10), "idle"), ((72, 16), "watch"), ((72, 23), "error"), ((72, 30), "stop")], [], NONE,
	      [((73, 3), ((73, 8), "idle"),
		StSequence ((74, 16), StWait ((74, 6), Closed d1, Closed d1), StTo ((75, 6), "watch"))),
	       ((76, 3), ((76, 8), "watch"),
		StSelect
		    ([[((79, 6), StSequence ((78, 4), StSignal (((78, 2), "E1"), ref Local, ref ([IN, OUT], [])), StTo ((78, 6), "stop")))],
		      [((77, 6), StSequence ((80, 14), StWait ((80, 2), Open (bdiff d2 d1), Infinite), StTo ((81, 2), "error")))]],
		     (82, 6)))]),
	 ComponentDecl
	     ((84, 0), ((84, 10), `"PresentWithin"), [], [], [],
	      [(([((85, 8), "E1")], ([IN, OUT], [])), SOME e, (Closed 0.0, Closed 0.0))],
	      [], NONE,
	      ParComp
		  ((86, 3), SomePorts [],
		   [(SomePorts [], InstanceComp (((86, 7), `"presentWithin"), ["E1"], []))])),
	 Main
	     (((88, 0), `"PresentWithin"), roots e),
	 PropertyDecl
	     ((90, 8), ((90, 17), `"present_within"),
	      LtlProp (BoxProp (NotProp (ProbeProp (StateProbe ((`"PresentWithin", [1]), "error"))))))],
	VarProp (L,`"present_within"))
    end

  | newPresentWithin _ =
    comperror "unsupported time interval in present .. within ..., sorry";


(*
process presentLasting [E1 : sync, notE1 : sync] is 
   states idle, firstE1, error, stop
   from idle
      E1;
      to firstE1
   from firstE1
      select
	 wait [D,D]; to stop
      unless
	 notE1 ; to error
      end
component PresentLasting is
   port E1 : sync in [0,0] is ...,
        notE1 : sync in [0,0] is not ...
   par presentLasting [E1, notE1] end
PresentLasting > ...
property present_lasting is ltl [] - (PresentLasting/1/state error)
*)

fun newPresentLasting (e,Closed d) =
    let val no = Int.toString (nextobs ())
	fun ` s = "_" ^ s ^ "_" ^ no
    in ([ProcessDecl
	     ((95, 0), ((95, 8), `"presentLasting"), [([((95, 24), "E1")], ([IN, OUT], [])), ([((95, 35), "notE1")], ([IN, OUT], []))], [],
	      [((96, 10), "idle"), ((96, 16), "firstE1"), ((96, 25), "error"), ((96, 32), "stop")], [], NONE,
	      [((97, 3), ((97, 8), "idle"),
		StSequence ((98, 8), StSignal (((98, 6), "E1"), ref Local, ref ([IN, OUT], [])), StTo ((99, 6), "firstE1"))),
	       ((100, 3), ((100, 8), "firstE1"),
		StSelect
		    ([[((103, 6), StSequence ((102, 7), StSignal (((102, 2), "notE1"), ref Local, ref ([IN, OUT], [])), StTo ((102, 9), "error")))],
		      [((101, 6), StSequence ((104, 12), StWait ((104, 2), Closed d, Closed d), StTo ((104, 14), "stop")))]],
		     (105, 6)))]),
	 ComponentDecl
	     ((107, 0), ((107, 10), `"PresentLasting"), [], [], [],
	      [(([((108, 8), "E1")], ([IN, OUT], [])), SOME e, (Closed 0.0, Closed 0.0)),
	       (([((109, 8), "notE1")], ([IN, OUT], [])), SOME (NotProbe e), (Closed 0.0, Closed 0.0))], [], NONE,
	      ParComp
		  ((110, 3), SomePorts [],
		   [(SomePorts [], InstanceComp (((110, 7), `"presentLasting"), ["E1", "notE1"], []))])),
	 Main
	     (((112, 0), `"PresentLasting"), roots e),
	 PropertyDecl
	     ((114, 8), ((114, 17), `"present_lasting"),
	      LtlProp (BoxProp (NotProp (ProbeProp (StateProbe ((`"PresentLasting", [1]), "error"))))))],
	VarProp (L,`"present_lasting"))
    end

  | newPresentLasting _ =
    comperror "unsupported time bound in present .. lasting ..., sorry";


(*
process absentAfter [E1 ,E2 : sync] is
   states idle, start, watch, error, ok
   from idle
      E2;
      to start
   from start
      wait [d1,d1];
      to watch
   from watch
      select
	 E1 ; to error
      unless
	 wait ]d2−d1,...[; to ok
      end
component AbsentAfter is
   port E1 : sync in [0,0] is ...,
        E2 : sync in [0,0] is ...
   par absentAfter [E1,E2] end
AbsentAfter > ...
property absent_after is ltl [] - (AbsentAfter/1/state error)
*)

fun newAbsentAfter (e1,e2,SOME (Closed d1,Closed d2)) =
    let val no = Int.toString (nextobs ())
	fun ` s = "_" ^ s ^ "_" ^ no
    in ([ProcessDecl
	     ((120, 0), ((120, 8), `"absentAfter"),
	      [([((120, 21), "E1"), ((120, 25), "E2")], ([IN, OUT], []))], [],
	      [((121, 10), "idle"), ((121, 16), "start"), ((121, 23), "watch"), ((121, 30), "error"), ((121, 37), "ok")], [], NONE,
	      [((122, 3), ((122, 8), "idle"),
		StSequence ((123, 8), StSignal (((123, 6), "E2"), ref Local, ref ([IN, OUT], [])), StTo ((124, 6), "start"))),
	       ((125, 3), ((125, 8), "start"),
		StSequence ((126, 16), StWait ((126, 6), Closed d1, Closed d1), StTo ((127, 6), "watch"))),
	       ((128, 3), ((128, 8), "watch"),
		StSelect
		    ([[((131, 6), StSequence ((130, 5), StSignal (((130, 2), "E1"), ref Local, ref ([IN, OUT], [])), StTo ((130, 7), "error")))],
		      [((129, 6),  StSequence ((132, 14), StWait ((132, 2), Open (bdiff d2 d1), Infinite), StTo ((132, 16), "ok")))]],
		     (133, 6)))]),
	 ComponentDecl
	     ((135, 0), ((135, 10), `"AbsentAfter"), [], [], [],
	      [(([((136, 8), "E1")], ([IN, OUT], [])), SOME e1, (Closed 0.0, Closed 0.0)),
	       (([((137, 8), "E2")], ([IN, OUT], [])), SOME e2,  (Closed 0.0, Closed 0.0))], [], NONE,
	      ParComp
		  ((138, 3), SomePorts [],
		   [(SomePorts [], InstanceComp (((138, 7), `"absentAfter"), ["E1", "E2"], []))])),
	 Main
	     (((140, 0), `"AbsentAfter"), roots (OrProbe(e1,e2))),
	 PropertyDecl
	     ((142, 8), ((142, 17), `"absent_after"),
	      LtlProp (BoxProp (NotProp (ProbeProp (StateProbe ((`"AbsentAfter", [1]), "error"))))))],
	VarProp (L,`"absent_after"))
    end

  | newAbsentAfter _ =
    comperror "unsupported time interval in absent .. after ... within ..., sorry";


(*
process absentBefore1 [E1, E3 : sync] (&bad : bool, &foundE2 : bool) is
   states idle
   from idle
      select
	 E3;
         on bad;
	 bad := false; to idle
      [] E1;
         on not bad;
	 bad := true ;
	 foundE2:= false; to idle
      end
process absentBefore2 [E2 : sync] (&bad : bool, &foundE2 : bool) is
   states idle, stop
   from idle
      E2;
      foundE2 := true;
      to stop
component AbsentBefore is
   var bad : bool := false,
       foundE2 : bool := false
   port E1 : sync in [0,0] /* is ... */,
        E2 : sync in [0,0] /* is ... */,
        E3 : sync in [2,2]                    // [d,d]
   priority E1 > E3
   par absentBefore1 [E1,E3] (&bad,&foundE2)
   ||  absentBefore2 [E2] (&bad,&foundE2)
   end
AbsentBefore // > ...
property absent_before is ltl (not <> (AbsentBefore/1/value foundE2 and AbsentBefore/1/value bad))
*)

fun newAbsentBefore (e1,e2,SOME (Closed d)) =
    let val no = Int.toString (nextobs ())
	fun ` s = "_" ^ s ^ "_" ^ no
    in ([ProcessDecl
	     ((152, 0), ((152, 8), `"absentBefore1"), [([((152, 23), "E1"), ((152, 27), "E3")], ([IN, OUT], []))],
	      [([RefArg ((152, 40), "bad")], [], BoolType), ([RefArg ((152, 53), "foundE2")], [], BoolType)],
	      [((153, 10), "idle")], [], NONE,
	      [((154, 3), ((154, 8), "idle"),
		StSelect
		    ([[((159, 6), StSequence ((156, 4), StSignal (((156, 2), "E3"), ref Local, ref ([IN, OUT], [])),
					      StSequence ((157, 8), StOn ((157, 2), IdentExp ("bad", ref (VarKind, BoolType))),
							  StSequence ((158, 14), StAssign ((158, 6), [IdentExp ("bad", ref (VarKind, BoolType))],[BoolExp false], ref true),
								      StTo ((158, 16), "idle"))))),
		       ((155, 6), StSequence ((159, 11), StSignal (((159, 9), "E1"), ref Local, ref ([IN, OUT], [])),
					      StSequence ((160, 12), StOn ((160, 2), UnopExp (NOT, ref BoolType, IdentExp ("bad", ref (VarKind, BoolType)))),
							  StSequence ((161, 14), StAssign ((161, 6), [IdentExp ("bad", ref (VarKind, BoolType))], [BoolExp true], ref true),
								      StSequence ((162, 17), StAssign ((162, 9), [IdentExp ("foundE2", ref (VarKind, BoolType))], [BoolExp false], ref true),
										  StTo ((162, 19), "idle"))))))]],
		     (163, 6)))]),
	 ProcessDecl
	     ((165, 0), ((165, 8), `"absentBefore2"), [([((165, 23), "E2")], ([IN, OUT], []))],
	      [([RefArg ((165, 36), "bad")], [], BoolType), ([RefArg ((165, 49), "foundE2")], [], BoolType)],
	      [((166, 10), "idle"), ((166, 16), "stop")], [], NONE,
	      [((167, 3), ((167, 8), "idle"),
		StSequence ((168, 8), StSignal (((168, 6), "E2"), ref Local, ref ([IN, OUT], [])),
			    StSequence ((169, 21), StAssign ((169, 14), [IdentExp ("foundE2", ref (VarKind, BoolType))], [BoolExp true], ref true),
					StTo ((170, 6), "stop"))))]),
	 ComponentDecl
	     ((172, 0), ((172, 10), `"AbsentBefore"), [], [],
	      [([((173, 7), "bad")], BoolType, SOME (BoolExp false)), ([((174, 7), "foundE2")], BoolType, SOME (BoolExp false))],
	      [(([((175, 8), "E1")], ([IN, OUT], [])), SOME e1, (Closed 0.0, Closed 0.0)),
	       (([((176, 8), "E2")], ([IN, OUT], [])), SOME e2, (Closed 0.0, Closed 0.0)),
	       (([((177, 8), "E3")], ([IN, OUT], [])), NONE,  (Closed d, Closed d))],
	      [([((178, 12), "E1")], [((178, 17), "E3")])], NONE,
	      ParComp
		  ((179, 3), SomePorts [],
		   [(SomePorts [], InstanceComp (((179, 7), `"absentBefore1"), ["E1", "E3"],
						 [RefExp ("bad", ref BoolType), RefExp ("foundE2", ref BoolType)])),
		    (SomePorts [], InstanceComp (((180, 7), `"absentBefore2"), ["E2"],
						 [RefExp ("bad", ref BoolType), RefExp ("foundE2", ref BoolType)]))])),
	 Main
	     (((183, 0), `"AbsentBefore"), roots (OrProbe (e1,e2))),
	 PropertyDecl
	     ((185, 8), ((185, 17), `"absent_before"),
	      LtlProp (NotProp (DiaProp (AndProp
					     (ProbeProp (ValueProbe ((`"AbsentBefore", [1]), IdentExp ("foundE2", ref (VarKind, BoolType)))),
					      ProbeProp  (ValueProbe ((`"AbsentBefore", [1]), IdentExp ("bad", ref (VarKind, BoolType)))))))))],
	VarProp (L,`"absent_before"))
    end

  | newAbsentBefore _ =
    comperror "unsupported time bound in absent .. before ... for ..., sorry";


(*
process absentWithin [E1 : sync] is
   states idle, watch, error, ok
   from idle
      wait [d1,d1];
      to watch
   from watch
      select
	 E1; to error
      unless
	 wait ]d2−d1,...[; to ok
      end
component AbsentWithin is
   port E1 : sync in [0,0] is ...
   par absentWithin [E1] end
AbsentWithin > ...
property absent_within is ltl [] - AbsentWithin/1/state error      
*)

fun newAbsentWithin (e,SOME (Closed d1,Closed d2)) =
    let val no = Int.toString (nextobs ())
	fun ` s = "_" ^ s ^ "_" ^ no
    in ([ProcessDecl
	     ((192, 0), ((192, 8), `"absentWithin"), [([((192, 22), "E1")], ([IN, OUT], []))], [],
	      [((193, 10), "idle"), ((193, 16), "watch"), ((193, 23), "error"), ((193, 30), "ok")], [], NONE,
	      [((194, 3), ((194, 8), "idle"),
		StSequence ((195, 16), StWait ((195, 6), Closed d1, Closed d1), StTo ((196, 6), "watch"))),
	       ((197, 3), ((197, 8), "watch"),
		StSelect
		    ([[((200, 6), StSequence ((199, 4), StSignal (((199, 2), "E1"), ref Local, ref ([IN, OUT], [])), StTo ((199, 6), "error")))],
		      [((198, 6), StSequence ((201, 14), StWait ((201, 2), Open (bdiff d2 d1), Infinite), StTo ((201, 16), "ok")))]],
		     (202, 6)))]),
	 ComponentDecl
	     ((204, 0), ((204, 10), `"AbsentWithin"), [], [], [],
	      [(([((205, 8), "E1")], ([IN, OUT], [])), SOME e, (Closed 0.0, Closed 0.0))], [], NONE,
	      ParComp
		  ((206, 3), SomePorts [],
		   [(SomePorts [], InstanceComp (((206, 7), `"absentWithin"), ["E1"], []))])),
	 Main
	     (((209, 0), `"AbsentWithin"), roots e),
	 PropertyDecl
	     ((211, 8), ((211, 17), `"absent_within"),
	      LtlProp (BoxProp (NotProp (ProbeProp (StateProbe ((`"AbsentWithin", [1]), "error"))))))],
	VarProp (L,`"absent_within"))
    end

  | newAbsentWithin _ =
    comperror "unsupported time interval in absent .. within ..., sorry";


(*
process leadstoWithin [E1, E2 : sync] is
   states idle, start, watch, error
   from idle
      E1;
      to start
   from start
      wait [d1,d1];
      to watch
   from watch
      select
	 E2; to idle
      unless
	 wait ]d2−d1,...[; to error
      end
component LeadstoWithin is
   port E1 : sync in [0,0] is ...,
        E2 : sync in [0,0] is ...
   par leadstoWithin [E1,E2] end
LeadstoWithin > ...
property leadsto_within is ltl [] - LeadstoWithin/1/state error
*)

fun newLeadstoWithin (e1,e2,SOME (Closed d1,Closed d2)) =
    let val no = Int.toString (nextobs ())
	fun ` s = "_" ^ s ^ "_" ^ no
    in ([ProcessDecl
	     ((216, 0), ((216, 8), `"leadstoWithin"), [([((216, 23), "E1"), ((216, 27), "E2")], ([IN, OUT], []))], [],
	      [((217, 10), "idle"), ((217, 16), "start"), ((217, 23), "watch"), ((217, 30), "error")], [], NONE,
	      [((218, 3), ((218, 8), "idle"),
		StSequence ((219, 8), StSignal (((219, 6), "E1"), ref Local, ref ([IN, OUT], [])), StTo ((220, 6), "start"))),
	       ((221, 3), ((221, 8), "start"),
		StSequence ((222, 16), StWait ((222, 6), Closed d1, Closed d1), StTo ((223, 6), "watch"))),
	       ((224, 3), ((224, 8), "watch"),
		StSelect
		    ([[((227, 6), StSequence ((226, 4), StSignal (((226, 2), "E2"), ref Local, ref ([IN, OUT], [])), StTo ((226, 6), "idle")))],
		      [((225, 6), StSequence ((228, 14), StWait ((228, 2), Open (bdiff d2 d1), Infinite), StTo ((228, 16), "error")))]],
		     (229, 6)))]),
	 ComponentDecl
	     ((231, 0), ((231, 10), `"LeadstoWithin"), [], [], [],
	      [(([((232, 8), "E1")], ([IN, OUT], [])), SOME e1, (Closed 0.0, Closed 0.0)),
	       (([((233, 8), "E2")], ([IN, OUT], [])), SOME e2, (Closed 0.0, Closed 0.0))], [], NONE,
	      ParComp
		  ((234, 3), SomePorts [],
		   [(SomePorts [], InstanceComp (((234, 7), `"leadstoWithin"), ["E1", "E2"], []))])),
	 Main
	     (((236, 0), `"LeadstoWithin"), roots (OrProbe (e1,e2))),
	 PropertyDecl
	     ((238, 8), ((238, 17), `"leadsto_within"),
	      LtlProp (BoxProp (NotProp (ProbeProp (StateProbe ((`"LeadstoWithin", [1]), "error"))))))],
	VarProp (L,`"leadsto_within"))
    end

  | newLeadstoWithin _ =
    comperror "unsupported time interval in ... leadsto ... within ..., sorry";


(*
process leadstoBefore1 [E1 , E2 : sync] (&foundE3 : bool) is
   states idle, start, watch, error
   from idle
      E1;
      on not foundE3;
      to start
   from start
      wait [d1,d1];
      to watch
   from watch
      select
	 E2; on not foundE3; to idle
      unless
	 wait ]d2−d1,...[; to error
      end
process leadstoBefore2 [E3 : sync] (&foundE3 : bool) is
   states idle, stop
   from idle
      E3;
      foundE3 := true;
      to stop
component LeadstoBefore is
   var foundE3 : bool := false
   port E1 : sync in [0,0] is ...,
        E2 : sync in [0,0] is ...,
        E3 : sync in [0,0] is ...
   par
      leadstoBefore1 [E1,E2] (&foundE3)
   || leadstoBerore2 [E3] (&foundE3)
   end

LeadstoBefore > ...

property leadsto_before is ltl <> (LeadstoBefore/event E3) => [] - LeadstoBefore/1/state error
*)

fun newLeadstoBefore (e1,e2,e3,SOME (Closed d1,Closed d2)) =
    let val no = Int.toString (nextobs ())
	fun ` s = "_" ^ s ^ "_" ^ no
    in ([ProcessDecl
	     ((245, 0), ((245, 8), `"leadstoBefore1"), [([((245, 24), "E1"), ((245, 29), "E2")], ([IN, OUT], []))],
	      [([RefArg ((245, 42), "foundE3")], [], BoolType)],
	      [((246, 10), "idle"), ((246, 16), "start"), ((246, 23), "watch"), ((246, 30), "error")], [], NONE,
	      [((247, 3), ((247, 8), "idle"),
		StSequence ((248, 8), StSignal (((248, 6), "E1"), ref Local, ref ([IN, OUT], [])),
			    StSequence ((249, 20), StOn ((249, 6), UnopExp (NOT, ref BoolType, IdentExp ("foundE3", ref (VarKind, BoolType)))),
					StTo ((250, 6), "start")))),
	       ((251, 3), ((251, 8), "start"),
		StSequence ((252, 16), StWait ((252, 6), Closed d1, Closed d1), StTo ((253, 6), "watch"))),
	       ((254, 3), ((254, 8), "watch"),
		StSelect
		    ([[((257, 6), StSequence ((256, 4), StSignal (((256, 2), "E2"), ref Local, ref ([IN, OUT], [])),
					      StSequence ((256, 20), StOn ((256, 6), UnopExp (NOT, ref BoolType, IdentExp ("foundE3", ref (VarKind, BoolType)))),
							  StTo ((256, 22), "idle"))))],
		      [((255, 6), StSequence ((258, 14), StWait ((258, 2), Open (bdiff d2 d1), Infinite), StTo ((258, 16), "error")))]],
		     (259, 6)))]),
	 ProcessDecl
	     ((261, 0), ((261, 8), `"leadstoBefore2"), [([((261, 24), "E3")], ([IN, OUT], []))],
	      [([RefArg ((261, 37), "foundE3")], [], BoolType)],
	      [((262, 10), "idle"), ((262, 16), "stop")], [], NONE,
	      [((263, 3), ((263, 8), "idle"),
		StSequence ((264, 8), StSignal (((264, 6), "E3"), ref Local, ref ([IN, OUT], [])),
			    StSequence ((265, 21), StAssign ((265, 14), [IdentExp ("foundE3", ref (VarKind, BoolType))], [BoolExp true], ref true),
					StTo ((266, 6), "stop"))))]),
	 ComponentDecl
	     ((268, 0), ((268, 10), `"LeadstoBefore"), [], [],
	      [([((269, 7), "foundE3")], BoolType, SOME (BoolExp false))],
	      [(([((270, 8), "E1")], ([IN, OUT], [])), SOME e1, (Closed 0.0, Closed 0.0)),
	       (([((271, 8), "E2")], ([IN, OUT], [])), SOME e2, (Closed 0.0, Closed 0.0)),
	       (([((272, 8), "E3")], ([IN, OUT], [])), SOME e3, (Closed 0.0, Closed 0.0))],
	      [], NONE,
	      ParComp
		  ((273, 3), SomePorts [],
		   [(SomePorts [], InstanceComp (((274, 6), `"leadstoBefore1"), ["E1", "E2"], [RefExp ("foundE3", ref BoolType)])),
		    (SomePorts [], InstanceComp (((275, 6), `"leadstoBefore2"), ["E3"], [RefExp ("foundE3", ref BoolType)]))])),
	 Main
	     (((278, 0), `"LeadstoBefore"), roots (OrProbe(e1,OrProbe(e2,e3)))),
	 PropertyDecl
	     ((280, 8), ((280, 17), `"leadsto_before"),
	      LtlProp (ImpProp (DiaProp (ProbeProp (EventProbe ((`"LeadstoBefore", []), "E3"))),
				BoxProp (NotProp (ProbeProp (StateProbe ((`"LeadstoBefore", [1]), "error")))))))],
	VarProp (L,`"leadsto_before"))
    end

  | newLeadstoBefore _ =
    comperror "unsupported time interval in ... leadsto ... before ... within ..., sorry";

(*
process leadstoAfter [E1,E2,E3 : sync] is
   states idle, start, watch, beginObs, error
   from idle
      E3;
      to start
   from start
      E1;
      to beginobs
   from beginObs
      wait [d1,d1];
      to watch
   from watch
      select
	 E2; to idle
      unless
	 wait ]d2−d1,...[; to error
      end
component LeadstoAfter is
   port E1 : sync in [0,0] is ...,
        E2 : sync in [0,0] is ...,
        E3 : sync in [0,0] is ...
   par leadstoAfter [E1,E2,E3] end
LeadstoAfter > ...
property leadsto_after is ltl <> (LeadstoAfter/event E3) => [] - LeadstoAfter/1/state error
*)

fun newLeadstoAfter (e1,e2,e3,SOME (Closed d1,Closed d2)) =
    let val no = Int.toString (nextobs ())
	fun ` s = "_" ^ s ^ "_" ^ no
    in ([ProcessDecl
	     ((286, 0), ((286, 8), `"leadstoAfter"), [([((286, 22), "E1"), ((286, 25), "E2"), ((286, 28), "E3")], ([IN, OUT], []))], [],
	      [((287, 10), "idle"), ((287, 16), "start"), ((287, 23), "watch"), ((287, 30), "beginObs"), ((287, 40), "error")], [], NONE,
	      [((288, 3), ((288, 8), "idle"),
		StSequence ((289, 8), StSignal (((289, 6), "E3"), ref Local, ref ([IN, OUT], [])), StTo ((290, 6), "start"))),
	       ((291, 3), ((291, 8), "start"),
		StSequence ((292, 8), StSignal (((292, 6), "E1"), ref Local, ref ([IN, OUT], [])),
			    StTo ((293, 6), "beginObs"))),
	       ((294, 3), ((294, 8), "beginObs"),
		StSequence ((295, 16), StWait ((295, 6), Closed d1, Closed d1), StTo ((296, 6), "watch"))),
	       ((297, 3), ((297, 8), "watch"),
		StSelect
		    ([[((300, 6), StSequence ((299, 4), StSignal (((299, 2), "E2"), ref Local, ref ([IN, OUT], [])), StTo ((299, 6), "idle")))],
		      [((298, 6), StSequence ((301, 14), StWait ((301, 2), Open (bdiff d2 d1), Infinite), StTo ((301, 16), "error")))]],
		     (302, 6)))]),
	 ComponentDecl
	     ((304, 0), ((304, 10), `"LeadstoAfter"), [], [], [],
	      [(([((305, 8), "E1")], ([IN, OUT], [])), SOME e1, (Closed 0.0, Closed 0.0)),
	       (([((306, 8), "E2")], ([IN, OUT], [])), SOME e2, (Closed 0.0, Closed 0.0)),
	       (([((307, 8), "E3")], ([IN, OUT], [])), SOME e3, (Closed 0.0, Closed 0.0))], [], NONE,
	      ParComp
		  ((308, 3), SomePorts [],
		   [(SomePorts [], InstanceComp (((308, 7), `"leadstoAfter"), ["E1", "E2", "E3"], []))])),
	 Main
	     (((310, 0), `"LeadstoAfter"), roots (OrProbe(e1,OrProbe(e2,e3)))),
	 PropertyDecl
	     ((312, 8), ((312, 17), `"leadsto_after"),
	      LtlProp (ImpProp (DiaProp (ProbeProp (EventProbe ((`"LeadstoAfter", []), "E3"))),
				BoxProp (NotProp (ProbeProp (StateProbe ((`"LeadstoAfter", [1]), "error")))))))],
	VarProp (L,`"leadsto_after"))
    end

  | newLeadstoAfter _ =
    comperror "unsupported time interval in ... leadsto ... after ... within ..., sorry";



fun expandProperty p =
    let val DL = ref []  (* declarations to be added *)
	fun repl f args = if !OBSD then comperror "option -obsD not implemented yet, sorry"
			  else let val (dl,p') = f args
			       in DL := !DL @ dl; 
				  p'
			       end
	fun rexpand p =
	    case p of
	      (* traps legal timed patterns and replace them with observers and observer formula *)
		PresentProp (pb,i as SOME _)                 => repl newPresentWithin (pb,i)
	      | PresentLastingProp (pb,b)                    => repl newPresentLasting (pb,b)
	      | PresentAfterProp (pb1,pb2,i as SOME _)       => repl newPresentAfter (pb1,pb2,i)
	      | PresentBeforeProp (pb1,pb2,i as SOME _)      => repl newPresentBefore (pb1,pb2,i)
	      | PresentBetweenProp (pb1,pb2,pb3,SOME _)      => comperror "unsupported timed pattern yet (present ... between ... within ...), sorry"
	      | PresentUntilProp (pb1,pb2,pb3,SOME _)        => comperror "unsupported timed pattern yet (present ... until ... within ...), sorry"
	      (* *)
	      | AbsentProp (pb,i as SOME _)                  => repl newAbsentWithin (pb,i)
	      | AbsentLastingProp (pb,b)                     => comperror "unsupported timed pattern yet (absent ... lasting ...), sorry"
	      | AbsentAfterProp (pb1,pb2,i as SOME _)        => repl newAbsentAfter (pb1,pb2,i)
	      | AbsentBeforeProp (pb1,pb2,b as SOME _)       => repl newAbsentBefore (pb1,pb2,b)
	      | AbsentBetweenProp (pb1,pb2,pb3,SOME _)       => comperror "unsupported timed pattern yet (absent ... between ... within ...), sorry"
	      | AbsentUntilProp (pb1,pb2,pb3,SOME _)         => comperror "unsupported timed pattern yet (absent ... until ... within ...), sorry"
	      (* *)
	      | LeadstoProp (pb1,pb2,i as SOME _)            => repl newLeadstoWithin (pb1,pb2,i)
	      | LeadstoBeforeProp (pb1,pb2,pb3,i as SOME _)  => repl newLeadstoBefore (pb1,pb2,pb3,i)
	      | LeadstoAfterProp (pb1,pb2,pb3,i as SOME _)   => repl newLeadstoAfter (pb1,pb2,pb3,i)
	      | LeadstoBetweenProp (pb1,pb2,pb3,pb4,SOME _)  => comperror "unsupported timed pattern yet (... leadsto ... between ... within ...), sorry"
	      | LeadstoUntilProp (pb1,pb2,pb3,pb4,SOME _)    => comperror "unsupported timed pattern yet (... leadsto ... until ... within ...), sorry"
	      (* *)
	      | AlwaysProp (pb,SOME _)                       => comperror "unsupported timed pattern yet (always ... within ...), sorry"
	      | AlwaysLastingProp (pb,_)                     => comperror "unsupported timed pattern yet (always ... lasting ...), sorry"
 	      | AlwaysAfterProp (pb1,pb2,SOME _)             => comperror "unsupported timed pattern yet (always ... after ... for ...), sorry"
	      | AlwaysBeforeProp (pb1,pb2,SOME _)            => comperror "unsupported timed pattern yet (always ... before ... for ...), sorry"
	      | AlwaysBetweenProp (pb1,pb2,pb3,SOME _)       => comperror "unsupported timed pattern yet (always ... between ... within ...), sorry"
	      | AlwaysUntilProp (pb1,pb2,pb3,SOME _)         => comperror "unsupported timed pattern yet (always ... until ... within ...), sorry"
	      (* *)
	      | PrecedesProp (pb1,pb2,SOME _)                => comperror "unsupported timed pattern yet (... precedes ... for ...), sorry"
	      | PrecedesBeforeProp (pb1,pb2,pb3,SOME _)      => comperror "unsupported timed pattern yet (... precedes ... before ... for ...), sorry"
	      | PrecedesAfterProp (pb1,pb2,pb3,SOME _)       => comperror "unsupported timed pattern yet (... precedes ... after ... for ...), sorry"
	      | PrecedesBetweenProp (pb1,pb2,pb3,pb4,SOME _) => comperror "unsupported timed pattern yet (... precedes ... between ... for ...), sorry"
	      | PrecedesUntilProp (pb1,pb2,pb3,pb4,SOME _)   => comperror "unsupported timed pattern yet (... precedes ... until ... for ...), sorry"
	      (* *)
	      | AndProp (p1,p2) => AndProp (rexpand p1,rexpand p2)
	      | OrProp (p1,p2) => OrProp (rexpand p1,rexpand p2)
	      | ImpProp (p1,p2) => ImpProp (rexpand p1,rexpand p2)
	      | NotProp p => NotProp (rexpand p)
	      (* note: patterns cannot appear in ltl properties *)
	      | _ => p
	val p' = rexpand p
    in (!DL,p')
    end;

fun expandDecls (Program dl) =
    Program (flat (map (fn d as PropertyDecl (l,s,p) => let val (dl',p') = expandProperty p in dl' @ [PropertyDecl (l,s,p')] end | d => [d]) dl))
    handle BYE => raise BYE | _ => failwith "observers";


end; (* Observers *)
